"""Test tools/check.py."""

from pathlib import Path

import pytest

from slipbox.app import App, startup
from slipbox.build import build
from slipbox.dependencies import check_requirements
from slipbox.tools import check


def scan(app: App) -> None:
    """Run 'slipbox build --no-output'."""
    app.args["output"] = False
    build(app)


def test_print_sequence_empty(capsys: pytest.CaptureFixture[str]) -> None:
    """print_sequence must not print anything if sequence is empty.

    The result must be False (empty).
    """
    assert not check.print_sequence("error", [])
    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert not stderr


def test_print_sequence_not_empty(capsys: pytest.CaptureFixture[str]) -> None:
    """print_sequence must print header if sequence is not empty.

    The result must be True (non-empty)."""
    assert check.print_sequence("hello", ["world", "!"])
    stdout, stderr = capsys.readouterr()
    assert stdout
    assert "hello" in stdout
    assert "world" in stdout
    assert "!" in stdout
    assert not stderr


@pytest.mark.skipif(not check_requirements(startup({})),
                    reason="missing requirements")
class TestsWithRequirements:    # pylint: disable=R0201
    """Tests with external requirements (e.g. pandoc, graphviz, etc.)."""
    def test_invalid_links(self, app: App) -> None:
        """invalid_links must return note and invalid ID."""
        Path("test.md").write_text("# 0 Test\n[](#1)\n\n# 2 Test\n[](#0)\n")
        scan(app)

        result = list(check.invalid_links(app))
        assert result == [((0, "Test", "test.md"), 1)]

    def test_isolated_notes(self, app: App) -> None:
        """isolated_notes must return untagged notes only."""
        Path("test.md").write_text("""# 0 Foo

#test

[](#1)

# 1 Bar

# 2 Baz
""")
        scan(app)

        result = list(check.isolated_notes(app))
        assert result == [(2, "Baz", "test.md")]

    def test_unsourced_notes_empty_bibliography(self, app: App) -> None:
        """unsourced_notes must be empty if there is no bibliography."""
        Path("test.md").write_text("# 0 Test\n\nTest.\n")
        scan(app)

        result = list(check.unsourced_notes(app))
        assert not result

    def test_unsourced_notes(self, app: App) -> None:
        """unsourced_notes must include every note that has no citation."""
        app.config.bibliography = Path("test.bib")

        Path("test.bib").write_text("""
@book{test_2020,
    title = {Title},
    language = {English},
    author = {Author},
    year = {2020},
}
""")

        Path("test.md").write_text("""# 0 Foo

Foo.

# 1 Bar

Bar.

[@test_2020]
""")
        scan(app)

        result = list(check.unsourced_notes(app))
        assert result == [(0, "Foo", "test.md")]

    def test_check_notes_empty(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """check_notes must not output anything if there are no errors.

        The result must be True (no errors).
        """
        assert check.check_notes(app)
        stdout, stderr = capsys.readouterr()
        assert not stdout
        assert not stderr

    def test_check_notes(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """check_notes must output to stdout.

        The result must be False (has errors).
        """
        Path("test.md").write_text("# 0 Test\n[](#1)")
        scan(app)

        assert not check.check_notes(app)
        stdout, stderr = capsys.readouterr()
        assert stdout
        assert "Test" in stdout
        assert not stderr



"""Check slipbox notes."""

import typing as t

from ..app import App

_Note = t.Tuple[int, str, str]


def print_sequence(header: str, sequence: t.Iterable[str]) -> bool:
    """Print header and sequence of items if sequence is not empty.

    Return bool to indicate that sequence is non-empty.
    """
    empty = True
    for item in sequence:
        if empty:
            empty = False
            print(header)
        print(item)
    if not empty:
        print()
    return not empty


def invalid_links(app: App) -> t.Iterator[t.Tuple[_Note, int]]:
    """Generate notes that link to invalid ID."""
    sql = """
        SELECT DISTINCT id, title, filename, dest
        FROM Links JOIN Notes ON src = id
        WHERE dest NOT IN (
            SELECT id FROM Notes
        )
        ORDER BY id
    """
    for nid, title, filename, dest in app.database.execute(sql):
        yield (nid, title, filename), dest


def isolated_notes(app: App) -> t.Iterator[_Note]:
    """Generate isolated notes."""
    yield from app.database.execute("""
        SELECT DISTINCT id, title, filename FROM Notes
        WHERE id NOT IN (
            SELECT src FROM ValidLinks UNION SELECT dest FROM ValidLinks
        )
    """)


def unsourced_notes(app: App) -> t.Iterator[_Note]:
    """Generate notes that need citations (only if there's a bibliography)."""
    if app.config.bibliography is not None:
        yield from app.database.execute("""
            SELECT DISTINCT id, title, filename FROM Notes
            WHERE id NOT IN (
                SELECT note FROM Citations
            )
        """)


def check_notes(app: App) -> bool:
    """Check notes in slipbox.

    Returns false is errors are found.
    """
    def format_note(note: _Note) -> str:
        return f"  {note[0]}. {note[1]} in {note[2]!r}."

    def format_link(link: t.Tuple[_Note, int]) -> str:
        return f"  {link[0][0]}. {link[0][1]} in {link[0][2]!r} -> {link[1]}."

    _invalid_links = invalid_links(app)
    _isolated_notes = isolated_notes(app)
    _unsourced_notes = unsourced_notes(app)

    errors = [
        print_sequence("The following notes link to non-existent notes.",
                       map(format_link, _invalid_links)),
        print_sequence("The following notes are not connected to other notes.",
                       map(format_note, _isolated_notes)),
        print_sequence("The following notes have missing citations.",
                       map(format_note, _unsourced_notes)),
    ]
    return not any(errors)



"""Test app.py."""

import pytest

from slipbox.app import App, find_root


def test_find_root_in_current(app: App) -> None:
    """find_root must find .slipbox if it's in the current directory."""
    assert find_root() == app.root


def test_find_root_in_parent(
    app: App,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """find_root must find .slipbox if it's in a parent directory."""
    path = app.root/"foo"/"bar"/"baz"
    path.mkdir(parents=True)
    monkeypatch.chdir(path)

    assert not path.joinpath(".slipbox").exists()

    root = find_root()
    assert root is not None
    assert app.root == root
    assert root.joinpath(".slipbox").exists()


def test_find_root_none() -> None:
    """find_root should return None if slipbox has not been initialized."""
    assert find_root() is None



from pathlib import Path
import re
import setuptools


def get_version() -> str:
    """Get version string from slipbox/__init__.py."""
    pattern = re.compile(r'__version__ = "(.+)"')
    init = Path(__file__).with_name("slipbox")/"__init__.py"
    result = pattern.search(init.read_text())

    assert result
    return result.groups()[0]


if __name__ == "__main__":
    setuptools.setup(
        name="slipbox",
        version=get_version(),
        author="Levi Gruspe",
        author_email="mail.levig@gmail.com",
        description="Static site generator for Zettelkasten notes",
        long_description=Path("README.md").read_text(),
        long_description_content_type="text/markdown",
        url="https://github.com/lggruspe/slipbox",
        packages=setuptools.find_packages(),
        package_data={
            "slipbox": [
                "data/*",
                "data/*/*",
                "data/*/*/*",
                "data/*/*/*/*",
                "data/*/*/*/*/*",
                "data/*/*/*/*/*/*",
                "migrations/*.sql",
                "templates/*",
            ]
        },
        classifiers=[
            "Environment :: Console",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
        ],
        install_requires=[
            "networkx>=2.5.1",
            "pydot>=1.4.2",
            "pyquery>=1.4.3",
        ],
        python_requires=">=3.7",
        entry_points={
            "console_scripts": [
                "slipbox=slipbox.__main__:main"
            ]
        }
    )



"""Slipbox configuration."""

from configparser import ConfigParser
from dataclasses import dataclass
import os
from pathlib import Path
import typing as t


@dataclass
class Config:
    """Slipbox config class."""
    # [slipbox]
    output_directory = Path("public")
    title = "Slipbox"

    # [note-patterns]
    patterns = {
        "*.md": True,
        "*.markdown": True,
        "*.rst": True,
    }

    # [paths]
    pandoc = "pandoc"
    dot = "dot"

    # [pandoc-options]
    bibliography = None
    strip_comments = True

    def read(self, config: ConfigParser) -> None:
        """Update from configparser."""
        self.output_directory = Path(
            config.get(
                "slipbox",
                "output_directory",
                fallback=self.output_directory,
            ),
        )
        self.title = config.get("slipbox", "title", fallback=self.title)

        for key, _ in config.items("note-patterns"):
            try:
                val = config.getboolean("note-patterns", key,
                                        fallback=self.patterns[key])
            except KeyError:
                continue
            self.patterns[key] = val

        self.pandoc = config.get("paths", "pandoc", fallback=self.pandoc)
        self.dot = config.get("paths", "dot", fallback=self.dot)

        bibliography = config.get(
            "pandoc-options",
            "bibliography",
            fallback=None,
        )
        self.bibliography = (
            Path(bibliography) if bibliography else self.bibliography
        )
        self.strip_comments = config.getboolean(
            "pandoc-options",
            "strip-comments",
            fallback=self.strip_comments,
        )

    def read_file(self, path: Path) -> None:
        """Update config from file."""
        config = ConfigParser()
        config.read_string(path.read_text())
        self.read(config)

    def read_dict(self, dict_: t.Dict[str, t.Any]) -> None:
        """Update config from dict."""
        config = ConfigParser()
        config.read_dict(dict_)
        self.read(config)

    def read_env(self) -> None:
        """Update config from environment variables.

        NOTE Not all options can be set using environment variables.
        """
        self.pandoc = os.getenv("PANDOC", self.pandoc)
        self.dot = os.getenv("DOT", self.dot)

    def to_config_parser(self) -> ConfigParser:
        """Convert to ConfigParser object."""
        config = ConfigParser()

        sections = ["slipbox", "note-patterns", "paths", "pandoc-options"]
        for section in sections:
            if not config.has_section(section):
                config.add_section(section)

        config.set("slipbox", "output_directory",
                   str(self.output_directory.resolve()))
        config.set("slipbox", "title", self.title)

        for pattern, include in self.patterns.items():
            config.set("note-patterns", pattern,
                       "true" if include else "false")

        config.set("paths", "pandoc", self.pandoc)
        config.set("paths", "dot", self.dot)

        if self.bibliography is not None:
            config.set(
                "pandoc-options",
                "bibliography",
                str(self.bibliography.resolve()),
            )
        config.set(
            "pandoc-options",
            "strip-comments",
            "true" if self.strip_comments else "false",
        )
        return config

    def write(self, path: Path) -> None:
        """Write config to file."""
        with open(path, "w") as file:
            self.to_config_parser().write(file)


__all__ = ["Config"]
